"""empty message

Revision ID: 1582bc5883ac
Revises: 
Create Date: 2021-04-19 15:07:44.848437

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1582bc5883ac'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('user',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=64), nullable=True),
    sa.Column('email', sa.String(length=120), nullable=True),
    sa.Column('password_hash', sa.String(length=128), nullable=True),
    sa.Column('roles', sa.String(length=512), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_user_email'), 'user', ['email'], unique=False)
    op.create_index(op.f('ix_user_name'), 'user', ['name'], unique=True)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_user_name'), table_name='user')
    op.drop_index(op.f('ix_user_email'), table_name='user')
    op.drop_table('user')
    # ### end Alembic commands ###



from .app import db
from .app import login

from werkzeug.security import generate_password_hash, check_password_hash

from flask_login import UserMixin

from time import time
import jwt
import json
from .app import app

class User(db.Model, UserMixin):
    __tablename__ = 'user'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(64), index=True, unique=True)
    email = db.Column(db.String(120), index=True)
    password_hash = db.Column(db.String(128))
    roles = db.Column(db.String(512), default="")
    attributes = db.Column(db.String(512), default="{}", nullable=False)

    def __repr__(self):
        return self.name

    def set_password(self, password):
        self.password_hash = generate_password_hash(password)
 
    def check_password(self, password):
        return check_password_hash(self.password_hash, password)

    def get_reset_password_token(self, expires_in=600):
        return jwt.encode(
            {'reset_password': self.id, 'exp': time() + expires_in},
            app.config['SECRET_KEY'], algorithm='HS256') #.decode('utf-8')

    @staticmethod
    def verify_reset_password_token(token):
        try:
            id = jwt.decode(token, app.config['SECRET_KEY'],
                            algorithms=['HS256'])['reset_password']
        except:
            return
        return User.query.get(id)
    
    def get_roles(self):
        return list(set(["everyone"] + (self.roles if self.roles is not None else "").lower().split(',')))

    def get_attr(self, k, default=None):
        return json.loads(self.attributes).get(k, default)
    
    def set_attr(self, k, val):
        ob = json.loads(self.attributes)
        ob[k] = val
        self.attributes = json.dumps(ob)

        db.session.commit()

    @staticmethod
    def get(username):
        return User.query.filter_by(name=username).first()

    @staticmethod
    def all():
        return User.query.all()

@login.user_loader
def load_user(id):
    return User.query.get(int(id))




import os
basedir = os.getcwd() + '/state'

class Config(object):

    SQLALCHEMY_DATABASE_URI = f'sqlite:///{basedir}/webapp.db'
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    SECRET_KEY = 'FILL_THIS_IN'
    SMTP_SERVER = "smtp.gmail.com"
    SMTP_PORT = 587
    SMTP_USERNAME = "FILL_THIS_IN"
    SMTP_PASSWORD = "FILL_THIS_IN"


os.makedirs(basedir, exist_ok=True)


import os
from shlex import quote
import sys
import pathlib
import time

import pyfra.shell
import pyfra.remote

class Experiment:
    def __init__(self, experiment_name, experiment_server=None):
        self.experiment_name = experiment_name
        
        if experiment_server is not None:
            self.delegate(experiment_server)

    def delegate(self, rem):
        tmux_name = f"pyfra_delegated_{self.experiment_name}"
        if isinstance(rem, str): rem = self.remote(rem)
    
        if "PYFRA_DELEGATED" in os.environ:
            return
        def _attach_tmux():
            rem.sh(f"tmux a -t {quote(tmux_name)}", maxbuflen=0, forward_keys=True)
        
        try:
            _attach_tmux()
        except pyfra.shell.ShellException:
            # todo: figure out how to forward ssh keys securely
            # the problem with ssh -A is that your remote is screwed if your original client goes offline, which totally
            # defeats the purpose of doing this in the first place. also there's something extremely weird going on
            # with tmux that makes it so that ssh forward won't carry over to the tmux session, but you can
            # fix that with `eval $(tmux show-env -s |grep '^SSH_')` inside the tmux, but that doesn't work with send-keys, it only works with 
            # actually running it by hand in the tmux for some reason. I've tried running it multiple times using send-keys,
            # adding a delay, adding a dummy ssh 127.0.0.1 command in between just to get ssh to use the ssh-agent to auth, etc and it jsut won't work. 
            # I'm not sure why, and to save my sanity for now I'm just going to require adding the right ssh keys to the delegated server manually.

            env = rem.env(tmux_name)

            ignore = [] if not pathlib.Path(".pyfraignore").exists() else pathlib.Path(".pyfraignore").read_text().strip().splitlines()
            pyfra.shell.copy(pyfra.remote.local.path("."), env.path("."), into=False, exclude=ignore)

            env.sh(f"sudo apt install tmux -y; pip install -U git+https://github.com/EleutherAI/pyfra; pip install -r requirements.txt; tmux new-session -d -s {quote(tmux_name)}", ignore_errors=True)
            # cmd = f"{cmd} || ( eval $(tmux show-env -s |grep '^SSH_'); {cmd} )"
            cmd = f"pyenv shell {env.pyenv_version}" if env.pyenv_version is not None else ""
            cmd += f";[ -f env/bin/activate ] && . env/bin/activate; "
            cmd += "PYFRA_DELEGATED=1 python "+" ".join([quote(x) for x in sys.argv])

            env.sh(f"tmux send-keys -t {quote(tmux_name)} {quote(cmd)} Enter")
            
            _attach_tmux()

        sys.exit(0)



from smtplib import SMTP

from email.message import EmailMessage

from flask import render_template

from threading import Thread

def create_email_message(from_address, to_address, subject, body, htmlBody):
    msg = EmailMessage()
    msg['From'] = from_address
    msg['To'] = to_address
    msg['Subject'] = subject
    msg.set_content(body)
    msg.set_content(htmlBody, subtype='html')
    return msg

def send_email(config, to_address, subject, body, htmlBody):
    mailserver = SMTP(config["SMTP_SERVER"], config["SMTP_PORT"])
    mailserver.ehlo()
    mailserver.starttls()
    mailserver.login(config["SMTP_USERNAME"], config["SMTP_PASSWORD"])

    message = create_email_message(config["SMTP_USERNAME"], to_address, subject, body, htmlBody)
    mailserver.send_message(message)

    mailserver.quit()

def send_password_reset_email(config, targetEmail, targetName, token):
    
    subject = "[Eleuther Flask App] Reset Your Password"
    Thread(target=send_email, 
           args=(config, targetEmail, subject,
                 render_template('email/reset_password.txt',
                                 username=targetName, token=token),
                 render_template('email/reset_password.html',
                                 username=targetName, token=token))).start()


from __future__ import annotations
from contextlib import contextmanager
from hashlib import new
from typing import *

from best_download import handler

import pyfra.shell
from pyfra.setup import install_pyenv
from collections import namedtuple
import codecs
import pickle
import shlex
import os
import random
import json
import csv
from natsort import natsorted
import io
import pathlib
from yaspin import yaspin
import uuid
import abc
from functools import partial, wraps
from colorama import Fore, Style
import imohash
import hashlib

sentinel = object()


def _normalize_homedir(x):
    """ Essentially expanduser(path.join("~", x)) but remote-agnostic """
    if x[:2] == './':
        x = x[2:]
    if x[-2:] == '/.':
        x = x[:-2]
    
    x = x.replace('/./', '/')

    if '~/' in x:
        x = x.split('~/')[-1]
    
    if x[-2:] == '/~':
        return '~'
    
    # some special cases
    if x == '': return '~'
    if x == '.': return '~'
    if x == '~': return '~'
    if x == '/': return '/'
    
    if x[0] != '/' and x[:2] != '~/':
        x = '~/' + x
    
    if x[-1] == '/' and len(x) > 1: x = x[:-1]
    
    return x


class _ObjectEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, pyfra.remote.RemotePath):
            return obj.sha256sum()
        if hasattr(obj, "_to_json"):
            return obj._to_json()
        
        return super().default(obj)


def hash_obs(*args):
    jsonobj = json.dumps(args, sort_keys=True, cls=_ObjectEncoder)
    arghash = hashlib.sha256(jsonobj.encode()).hexdigest()
    return arghash


def _print_skip_msg(envname, fn, hash):
    print(f"{Style.BRIGHT}[{envname.ljust(15)} {Style.DIM}§{Style.RESET_ALL}{Style.BRIGHT}{fn.rjust(10)}]{Style.RESET_ALL} Skipping {hash}")


def mutates_state(fn, hash_key=None):
    """
    Decorator that marks a function as mutating the state of the underlying environment.
    """
    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        if self._no_hash: return fn(self, *args, **kwargs)
        new_hash = self.update_hash(fn.__name__, *args, **kwargs) if hash_key is None else self.update_hash(*hash_key(fn, *args, **kwargs))
        try:
            # if hash is in the state, then we can just return that
            ret = self.get_kv(new_hash)
            _print_skip_msg(self.envname, fn.__name__, new_hash)
            
            return ret
        except KeyError:
            # otherwise, we need to run the function and save the result
            ret = fn(self, *args, **kwargs)
            self.set_kv(new_hash, ret)
            return ret
    return wrapper


# remote stuff

# global cache
_remotepath_cache = {}
_remotepath_modified_time = {}
def cache(fn):
    """
    Use as an annotation. Caches the response of the function and
    check modification time on the file on every call.
    """
    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        modified_time = self.stat().st_mtime
        hash = hash_obs(fn.__name__, args, kwargs)
        if hash not in _remotepath_cache or modified_time != _remotepath_modified_time[hash]:
            ret = fn(self, *args, **kwargs)
            _remotepath_cache[(self.remote.ip, self.fname, hash)] = ret
            _remotepath_modified_time[(self.remote.ip, self.fname, hash)] = modified_time
            return ret
        else:
            return _remotepath_cache[hash]
    return wrapper


class RemotePath:
    """
    A RemotePath represents a path somewhere on some Remote. The RemotePath object can be used to manipulate the file.

    Example usage: ::

        # write text
        rem.path("goose.txt").write("honk")

        # read text
        print(rem.path("goose.txt").read())

        # write json
        rem.path("goose.json").jwrite({"honk": 1})

        # read json
        print(rem.path("goose.json").jread())

        # write csv
        rem.path("goose.csv").csvwrite([{"col1": 1, "col2": "duck"}, {"col1": 2, "col2": "goose"}])

        # read csv
        print(rem.path("goose.csv").csvread())

        # copy stuff to/from remotes
        copy(rem1.path('goose.txt'), 'test1.txt')
        copy('test1.txt', rem2.path('goose.txt'))
        copy(rem2.path('goose.txt'), rem1.path('testing123.txt'))
    """
    def __init__(self, remote, fname):
        if remote is None: remote = local

        self.remote = remote
        self.fname = fname
    
    def rsyncstr(self) -> str:
        return f"{self.remote.ip}:{self.fname}" if self.remote is not None and self.remote.ip is not None else self.fname

    def _to_json(self) -> Dict[str, str]:
        return {
            'remote': self.remote.ip if self.remote is not None else None,
            'fname': self.fname,
        }

    def __repr__(self) -> str:
        return f"RemotePath({json.dumps(self._to_json())})"
    
    def _set_cache(self, fn_name, value, *args, **kwargs):
        modified_time = self.stat().st_mtime
        hash = hash_obs(fn_name, args, kwargs)
        _remotepath_modified_time[(self.remote.ip, self.fname, hash)] = modified_time
        _remotepath_cache[(self.remote.ip, self.fname, hash)] = value

    def read(self) -> str:
        """
        Read the contents of this file into a string
        """
        if self.remote.is_local():
            with open(os.path.expanduser(self.fname)) as fh:
                return fh.read()
        else:
            # TODO: replace with paramiko
            nonce = random.randint(0, 99999)
            pyfra.shell.copy(self, f".tmp.{nonce}", quiet=True)
            with open(f".tmp.{nonce}") as fh:
                ret = fh.read()
                pyfra.shell.rm(f".tmp.{nonce}")
                return ret
    
    def write(self, content, append=False) -> str:
        """
        Write text to this file.

        Args:
            content (str): The text to write
            append (bool): Whether to append or overwrite the file contents

        """
        self.remote.fwrite(self.fname, content, append)
    
    def jread(self) -> Dict[str, Any]:
        """
        Read the contents of this json file and parses it. Equivalent to :code:`json.loads(self.read())`
        """
        return json.loads(self.read())

    def jwrite(self, content):
        """
        Write a json object to this file. Equivalent to :code:`self.write(json.dumps(content))`

        Args:
            content (json): The json object to write

        """
        self.write(json.dumps(content))

    def csvread(self, colnames=None) -> List[dict]:
        """
        Read the contents of this csv file and parses it into an array of dictionaries where the keys are column names.

        Args:
            colnames (list): Optionally specify the names of the columns for csvs without a header row.
        """
        fh = io.StringIO(self.read())
        if self.fname[-4:] == ".tsv":
            rdr = csv.reader(fh, delimiter="\t")
        else:
            rdr = csv.reader(fh)

        if colnames:
            cols = colnames
        else:
            cols = list(next(rdr))
        
        for ob in rdr:
            yield {
                k: v for k, v in zip(cols, [*ob, *[None for _ in range(len(cols) - len(ob))]])
            }
    
    def csvwrite(self, data, colnames=None):
        """
        Write a list of dicts object to this csv file.

        Args:
            content (List[dict]): A list of dicts where the keys are column names. Every dicts should have the exact same keys.

        """
        fh = io.StringIO()
        if colnames is None:
            colnames = data[0].keys()

        wtr = csv.writer(fh)
        wtr.writerow(colnames)

        for dat in data:
            assert dat.keys() == colnames

            wtr.writerow([dat[k] for k in colnames])
        
        fh.seek(0)
        self.write(fh.read())
    
    def _remote_payload(self, name, *args, **kwargs):
        """
        Run an arbitrary Path.* function remotely and return the result.
        Restricted to os rather than arbitrary eval for security reasons.
        """
        assert all(x in "_.abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ" for x in name)

        if self.remote.is_local():
            fn = pathlib.Path(self.fname)
            for k in name.split("."):
                fn = getattr(fn, k)
            return fn(*args, **kwargs)
        else:
            payload = f"import pathlib,json; print(json.dumps(pathlib.Path({repr(self.fname)}).expanduser().{name}(*{args}, **{kwargs})))"
            ret = self.remote.sh(f"python -c {payload | pyfra.shell.quote}", quiet=True, no_venv=True, pyenv_version=None)
            return json.loads(ret)

    def stat(self) -> os.stat_result:
        """
        Stat a remote file
        """
        return os.stat_result(self._remote_payload("stat"))
    
    def exists(self) -> bool:
        """
        Check if this file exists
        """
        try:
            return self._remote_payload("exists")
        except pyfra.shell.ShellException:
            # if we can't connect to the remote, the file does not exist
            return False
    
    def is_dir(self) -> bool:
        """
        Check if this file exists
        """
        return self._remote_payload("is_dir")
    
    def unlink(self) -> None:
        """
        Delete this file
        """
        self._remote_payload("unlink")

    def expanduser(self) -> RemotePath:
        """
        Return a copy of this path with the home directory expanded.
        """
        if self.remote.is_local():
            return RemotePath(None, os.path.expanduser(self.fname))
        else:
            homedir = self.remote.home()

            # todo: be more careful with this replace
            return RemotePath(self.remote, os.path.expanduser(self.fname).replace("~", homedir))
    
    def __div__(self, other):
        return RemotePath(self.remote, os.path.join(self.fname, other))

    @cache
    def sha256sum(self) -> str:
        """
        Return the sha256sum of this file.
        """
        return self.remote.sh(f"sha256sum {self.fname}", quiet=True).split(" ")[0]

    @cache
    def quick_hash(self) -> str:
        """
        Get a hash of this file that catches file changes most of the time
        by hashing blocks from the file at th beginning, middle, and end.
        Really useful for getting a quick hash of a really big file, but obviously
        unsuitable for guaranteeing file integrity.

        Uses imohash under the hood.
        """
        params = {
            "hexdigest": True,
            "sample_size": 4 * 1024**2, # 4 MB
            "sample_threshhold": 16 * 1024**2, # 16 MB
        }
        if self.remote.is_local():
            return imohash.hashfile(self.fname, **params)
        else:
            # TODO: use paramiko
            # TODO: make faster by not trying to install every time
            payload = f"import imohash,json,os; print(json.dumps(imohash.hashfile(os.path.expanduser({repr(self.fname)}), **{params})))"
            ret = self.remote.sh(f"[ -f ~/.imohash ] || ( python -m pip --help > /dev/null 2>&1 || sudo apt-get install python3-pip -y > /dev/null 2>&1; python -m pip install imohash > /dev/null 2>&1; touch ~/.imohash ); python -c {payload | pyfra.shell.quote}", no_venv=True, pyenv_version=None, quiet=True)
            return json.loads(ret)
            

class Remote:
    def __init__(self, ip=None, wd=None, experiment=None):
        """
        Args:
            ip (str): The host to ssh to. This looks something like :code:`12.34.56.78` or :code:`goose.com` or :code:`someuser@12.34.56.78` or :code:`someuser@goose.com`. You must enable passwordless ssh and have your ssh key added to the server first. If None, the Remote represents localhost.
            wd (str): The working directory on the server to start out on.
            python_version (str): The version of python to use (i.e running :code:`Remote("goose.com", python_version="3.8.10").sh("python --version")` will use python 3.8.10). If this version is not already installed, pyfra will install it.
        """
        if ip in ["127.0.0.1", "localhost"]: ip = None

        self.ip = ip
        self.wd = _normalize_homedir(wd) if wd is not None else "~"
        self.experiment = experiment

        self._home = None
        self._no_hash = True

    def env(self, envname, git=None, branch=None, python_version="3.9.4") -> Remote:
        """
        Args:
            envname (str): The name for the environment.
            git (str): A git repo to clone for the environment. If the repo already exists it (and any local changes) will be reset and overwritten with a fresh clone. If None, nothing will be cloned. The requirements.txt will be automatically installed in the virtualenv.
            branch (str): Check out a particular branch. Defaults to whatever the repo default is.
            python_version (str): The python version for this environment. Defaults to the python version of this Remote.
        """

        return Env(ip=self.ip, git=git, branch=branch, python_version=python_version, envname=envname)

    def sh(self, x, quiet=False, wrap=True, maxbuflen=1000000000, ignore_errors=False, no_venv=False, pyenv_version=None, forward_keys=False):
        """
        Run a series of bash commands on this remote. This command shares the same arguments as :func:`pyfra.shell.sh`.
        """
        if self.ip is None:
            return pyfra.shell.sh(x, quiet=quiet, wd=self.wd, wrap=wrap, maxbuflen=maxbuflen, ignore_errors=ignore_errors, no_venv=no_venv, pyenv_version=pyenv_version)
        else:
            return pyfra.shell._rsh(self.ip, x, quiet=quiet, wd=self.wd, wrap=wrap, maxbuflen=maxbuflen, ignore_errors=ignore_errors, no_venv=no_venv, pyenv_version=pyenv_version, forward_keys=forward_keys)

    def path(self, fname=None) -> RemotePath:
        """
        This is the main way to make a :class:`RemotePath` object; see RemotePath docs for more info on what they're used for.

        If fname is not specified, this command allocates a temporary path
        """
        if fname is None:
            return self.path(f"pyfra_tmp_{uuid.uuid4().hex}")

        if isinstance(fname, RemotePath):
            assert fname.remote == self
            return fname

        return RemotePath(self, _normalize_homedir(os.path.join(self.wd, fname) if self.wd is not None else fname))

    def __repr__(self):
        return (self.ip if self.ip is not None else "127.0.0.1") + ":" + self.wd

    def fingerprint(self):
        """
        A unique string for the server that this Remote is pointing to. 
        """
        self.sh("if [ ! -f ~/.pyfra.fingerprint ]; then cat /dev/urandom | tr -dc 'a-zA-Z0-9' | fold -w 32 | head -n 1 > ~/.pyfra.fingerprint; fi", quiet=True)
        tmpname = ".fingerprint." + str(random.randint(0, 99999))
        pyfra.shell.copy(self.path("~/.pyfra.fingerprint"), tmpname)
        ret = pyfra.shell.fread(tmpname)
        pyfra.shell.rm(tmpname)
        return ret.strip()
    
    def _to_json(self):
        return {
            'ip': self.ip,
            'wd': self.wd,
        }

    def ls(self, x='.'):
        return list(natsorted(self.sh(f"ls {x} | cat").strip().split("\n")))
     
    def home(self):
        if self._home is None:
            self._home = self.sh("echo $HOME", quiet=True).strip()
        
        return self._home

    def fwrite(self, fname, content, append=False):
        if self.ip is None:
            with open(os.path.expanduser(fname), 'a' if append else 'w') as fh:
                fh.write(content)
        else:
            nonce = random.randint(0, 99999)
            with open(f".tmp.{nonce}", 'w') as fh:
                fh.write(content)
            if append:
                pyfra.shell.copy(f".tmp.{nonce}", self.path(f".tmp.{nonce}"), quiet=True)
                self.sh(f"cat .tmp.{nonce} >> {fname} && rm .tmp.{nonce}", quiet=True)
            else:
                pyfra.shell.copy(f".tmp.{nonce}", self.path(fname), quiet=True)
            pyfra.shell.rm(f".tmp.{nonce}")

    # key-value store for convienence

    def set_kv(self, key: str, value: Any) -> None:
        """
        A key value store to keep track of stuff in this env. The data is stored in the env
        on the remote. Current implementation is inefficient (linear time, and the entire json 
        file is moved each time) but there shouldn't be a lot of data stored in it anyways
        (premature optimization is bad), and if we need to store a lot of data in the future 
        we can always make this more efficient without changing the interface.
        """
        with self.no_hash():
            # TODO: make more efficient
            statefile = self.path(".pyfra_env_state.json")
            if statefile.exists():
                ob = statefile.jread()
            else:
                ob = {}
            pickled_value = pickle.dumps(value, 0).decode()
            ob[key] = pickled_value
            statefile.jwrite(ob)

    def get_kv(self, key: str) -> Any:
        """
        Retrieve a value from the state file.
        """
        with self.no_hash():
            # TODO: make more efficient
            statefile = self.path(".pyfra_env_state.json")
            if statefile.exists():
                ob = statefile.jread()
            else:
                ob = {}

            return pickle.loads(ob[key].encode())

    def update_hash(self, *args, **kwargs):
        pass
    
    @contextmanager
    def no_hash(self):
        """ Context manager to turn off hashing temporarily """
        if self._no_hash:
            yield
            return

        self._no_hash = True
        try:
            yield
        finally:
            self._no_hash = False

    def is_local(self):
        return self.ip is None
    
    def __hash__(self):
        # this hash is not the Env hash, which represents the state inside the Env,
        # but rather is supposed to represent the particular location this 
        # Remote/Env is pointing to.
        return hash((self.ip, self.wd))

# env
class Env(Remote):
    """
    An environment is a Remote pointing to a directory that has a virtualenv and a specific version version of python installed, optionally initialized from a git repo. Since environments are also just Remotes, all methods on Remotes work on environments too (including env itself, which makes a nested environment within that invironment with no problems whatsoever).

    A typical design pattern sees functions accepting remotes as argument and immediately turning it into an env that's used for the rest of the function. 

    Example usage: ::

        def train_model(rem, ...):
            e = rem.env("neo_experiment", "https://github.com/EleutherAI/gpt-neo", python_version="3.8.10")
            e.sh("do something")
            e.sh("do something else")
            f = some_other_thing(e, ...)
            e.path("goose.txt").write(f.jread()["honk"])
        
        def some_other_thing(rem, ...):
            # this makes an env inside the other env
            e = rem.env("other_thing", "https://github.com/EleutherAI/thing-doer-5000", python_version="3.8.10")
            e.sh("do something")
            e.sh("do something else")

            return e.path("output.json")
    """
    def __init__(self, ip=None, git=None, branch=None, force_rerun=False, python_version="3.9.4", envname=None):
        self.wd = f"~/pyfra_envs/{envname}"
        super().__init__(ip, self.wd)
        self.pyenv_version = python_version

        self.envname = envname

        self.hash = self._hash(None)
        self._no_hash = False

        if force_rerun:
            self.path(".pyfra_env_state.json").unlink()

        self._init_env(git, branch, python_version)

        global_env_registry.register(self)
    
    @classmethod
    def _hash(cls, *args, **kwargs):
        return hash_obs([args, kwargs])

    @mutates_state
    def _init_env(self, git, branch, python_version):
        with yaspin(text="Loading", color="white") as spinner, self.no_hash():
            ip = self.ip
            wd = self.wd

            spinner.text = f"[{ip}:{wd}] Installing python in env" 
            # install python/pyenv
            with spinner.hidden():
                self._install(python_version)

            self.sh(f"mkdir -p {wd}", no_venv=True, quiet=True)

            # pull git
            if git is not None:
                spinner.text = f"[{ip}:{wd}] Cloning from git repo" 
                # TODO: make this usable
                nonce = str(random.randint(0, 99999))

                if branch is None:
                    branch_cmds = ""
                else:
                    branch_cmds = f"git checkout {branch}; git pull origin {branch}; "

                self.sh(f"{{ rm -rf ~/.tmp_git_repo.{nonce} ; git clone {git} ~/.tmp_git_repo.{nonce} ; rsync -ar --delete ~/.tmp_git_repo.{nonce}/ {wd}/ ; rm -rf ~/.tmp_git_repo.{nonce} ; cd {wd}; {branch_cmds} }}", ignore_errors=True, quiet=True)

            # install venv
            if wd is not None:
                spinner.text = f"[{ip}:{wd}] Creating virtualenv" 
                pyenv_cmds = f"[ -d env/lib/python{python_version.rsplit('.')[0]} ] || rm -rf env ; python --version ; pyenv shell {python_version} ; python --version;" if python_version is not None else ""
                self.sh(f"mkdir -p {wd}; cd {wd}; {pyenv_cmds} [ -f env/bin/activate ] || python -m virtualenv env || ( python -m pip install virtualenv; python -m virtualenv env )", no_venv=True, quiet=True)
                spinner.text = f"[{ip}:{wd}] Installing requirements" 
                self.sh("pip install -e . ; pip install -r requirements.txt", ignore_errors=True, quiet=True)
            
            spinner.text = f"[{ip}:{wd}] Env created" 
            spinner.color = "green"
            spinner.ok("OK ")

    @mutates_state
    def sh(self, x, quiet=False, wrap=True, maxbuflen=1000000000, ignore_errors=False, no_venv=False, pyenv_version=sentinel, forward_keys=False):
        """
        Run a series of bash commands on this remote. This command shares the same arguments as :func:`pyfra.shell.sh`.
        """
    
        return super().sh(x, quiet=quiet, wrap=wrap, maxbuflen=maxbuflen, ignore_errors=ignore_errors, no_venv=no_venv, pyenv_version=pyenv_version if pyenv_version is not sentinel else self.pyenv_version, forward_keys=forward_keys)

    def _install(self, python_version):   
        # set up remote python version
        if python_version is not None: install_pyenv(self, python_version)

        # install rsync for copying files
        self.sh("rsync --help > /dev/null || ( sudo apt-get update && sudo apt-get install -y rsync )", quiet=True)

    def _to_json(self):
        return {
            'ip': self.ip,
            'wd': self.wd,
            'pyenv_version': self.pyenv_version,
        }

    def fwrite(self, fname, content, append=False):
        # wraps fwrite to make it keep track of state hashes
        # TODO: replace with paramiko
        # TODO: extract this statehash code and the one in shell.copy to a common function or something
        needs_set_kv = False
        if not self._no_hash:
            assert fname.startswith(self.wd)
            fname_suffix = fname[len(self.wd):]
            new_hash = self.update_hash("fwrite", fname_suffix, content, append)
            try:
                self.get_kv(new_hash)
                _print_skip_msg(self.envname, "fwrite", new_hash)
                return
            except KeyError:
                needs_set_kv = True
        
        with self.no_hash():
            super().fwrite(fname, content, append)

        if needs_set_kv:
            self.set_kv(new_hash, None)

    def update_hash(self, *args, **kwargs):
        self.hash = self._hash(self.hash, *args, **kwargs)
        return self.hash


def block(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        # get all Envs in args and kwargs
        envs = [(i, x) for i, x in enumerate(args) if isinstance(x, Env)] + \
            [(k, kwargs[k]) for k in sorted(kwargs.keys()) if isinstance(kwargs[k], Env)]
        inp_hashes = {
            k: v.hash for k, v in envs
        }

        def _prepare_for_hash(ind):
            # we want to handle Envs and RemotePaths specially:
            # for Envs, we only care about the Env hash
            # for RemotePaths, we only car about the quick_hash
            if ind in inp_hashes:
                return inp_hashes[ind]
            elif isinstance(ind, int) and isinstance(args[ind], RemotePath):
                return args[ind].quick_hash()
            elif isinstance(ind, str) and isinstance(kwargs[ind], RemotePath):
                return kwargs[ind].quick_hash()
            elif isinstance(ind, int): # normal arg type
                return args[ind]
            elif isinstance(ind, str):
                return kwargs[ind]
            else:
                raise Exception(f"Unknown ind type: {type(ind)}")

        # get a hash of all the inputs, except Env objects are counted as their hashes rather than the actual ip and wd
        overall_input_hash = hash_obs(
            [_prepare_for_hash(i) for i in range(len(args))],
            [_prepare_for_hash(k) for k in sorted(kwargs.keys())],
        )

        global_hashes_before = global_env_registry.hashes_by_env()

        try:
            # todo: detect RemotePaths in return value and substitute if broken
            # todo: handle Env objects that change the server they're on

            # the following different cases of envs passed are possible:
            # - created in block and not returned:
            #     as long as the Env is never 
            #     independently created again later, we don't need to do anything 
            #     special to track it; if it is created again, we need some kind
            #     of global tracking to tell it where to resume to
            # - created in block and returned
            #     comes for free because we save the return values. there is
            #     the problem that if the original Env disappears we might need to 
            #     rerun it, but we can figure that out later
            # - passed from outside block and not returned
            #     we Want to set the hash of the env, because it might be used 
            #     elsewhere by the caller. global tracking would also be useful here
            # - passed from outside block and returned
            #     same as last case, and we can basically ignore the return
            # - global from outside block and not returned
            #     it would be bad if we skip this block but don't update the hash of 
            #     the env, because it might be used elsewhere by the caller, and then 
            #     the hash will be all wrong.
            # - global from outside block and returned
            #     same as last case, except it's slightly easier to detect since we 
            #     can parse the output

            # set hashes for envs
            changed_hashes, ret = local.get_kv(overall_input_hash)

            envs_by_ip_envname = global_env_registry.envs_by_ip_envname()

            for ip, envname, orighash, newhash in changed_hashes:
                env = envs_by_ip_envname[(ip, envname)]
                if env.hash != orighash:
                    print(f"WARNING: expected env {ip}:{envname} to have hash {orighash} but got {env.hash}! Did the ip change?")
                env.hash = newhash
                
            print(f"Skipping block {fn.__name__}")
        except KeyError:
            ret = fn(*args, **kwargs)

            global_hashes_after = global_env_registry.hashes_by_env()

            # get the hashes that changed
            changed_hashes = [
                (k.ip, k.envname, global_hashes_before[k], global_hashes_after[k]) for k in global_hashes_after if global_hashes_before[k] != global_hashes_after[k]
            ]

            local.set_kv(overall_input_hash, (changed_hashes, ret))
        return ret
    return wrapper


class EnvRegistry:
    # everything here is O(n) but there shouldn't be a lot of envs so it's fine
    def __init__(self):
        self.envs = []
    
    def hashes_by_env(self):
        return {
            v: v.hash for v in self.envs
        }
    
    def envs_by_ip_envname(self):
        return {
            (v.ip, v.envname): v for v in self.envs
        }

    def register(self, env):
        self.envs.append(env)


local = Remote(wd=os.getcwd())
global_env_registry = EnvRegistry()


from .server import *
from functools import wraps, partial
import inspect
import string
import random
import os
from html import escape
try:
    from typing_extensions import Literal
except ModuleNotFoundError:
    from typing import Literal

from flask_migrate import upgrade
from flask_wtf import FlaskForm
from wtforms import StringField, BooleanField, SubmitField, IntegerField, SelectField
from wtforms.validators import DataRequired, Email, EqualTo
from ansi2html import Ansi2HTMLConverter


__all__ = [
    'page', 'webserver', 
    'current_user', # so we can know who's making each request inside a @page annotated function
    'User'
]


def dict_replace_if_fn(d):
    return {
        k: v() if callable(v) else v
        for k, v in d.items()
    }


def page(pretty_name=None, display: Literal["raw", "text", "monospace"]="monospace", field_names={}, dropdowns={}, roles=['everyone']):
    def _fn(callback, pretty_name, field_names, roles, display, dropdowns):

        sig = inspect.signature(callback)

        def make_form_class(dropdowns):
            dropdowns = dict_replace_if_fn(dropdowns)
            class CustomForm(FlaskForm):
                pass

            for name in sig.parameters:
                type = sig.parameters[name].annotation
                is_required = (sig.parameters[name].default == inspect._empty)

                if type == int:
                    field = IntegerField
                elif type == bool:
                    field = BooleanField
                else:
                    if name in dropdowns:
                        field = partial(SelectField, choices=dropdowns[name])
                    else:
                        field = StringField

                setattr(CustomForm, name, field(
                    field_names.get(name, name), 
                    validators=[DataRequired()] if is_required else [],
                    default = sig.parameters[name].default if not is_required else None
                    ))

            if len(sig.parameters) > 0:
                CustomForm.submit = SubmitField('Submit')
            return CustomForm

        form = len(sig.parameters) > 0

        def _callback_wrapper(k):
            html = callback(**k)
            if display == "raw":
                pass
            elif display == "text":
                html = escape(html)
            elif display == "monospace":
                converter = Ansi2HTMLConverter()
                html = converter.convert(html, full=False)
                html = f"<span class=\"monospace\">{html}</span>"
                html += converter.produce_headers()
            else:
                raise NotImplementedError
            
            return html

        register_page(callback.__name__, pretty_name, partial(make_form_class, dropdowns), _callback_wrapper, roles, redirect_index=False, has_form=form)
    
    # used @form and not @form()
    if callable(pretty_name):
        return _fn(pretty_name, pretty_name=None, field_names=field_names, roles=roles, display=display, dropdowns=dropdowns)

    return partial(_fn, pretty_name=pretty_name, field_names=field_names, roles=roles, display=display, dropdowns=dropdowns)


def gen_pass(stringLength=16):
    """Generate a random string of letters, digits """
    password_characters = string.ascii_letters + string.digits
    return ''.join(random.choice(password_characters) for i in range(stringLength))   


@page("Add User", roles=["admin"])
def adduser(username: str, email: str="example@example.com", roles: str=""):
    password = gen_pass()

    add_user(username, email, password, roles)

    return f"Added user {username} with randomly generated password {password}."

@page("Reset User Password", roles=["admin"])
def set_password(username: str):
    password = gen_pass()

    user = User.query.filter_by(name=username).first()

    user.set_password(password)
    db.session.commit()

    return f"Updated user {username} with randomly generated password {password}."


def webserver(debug=False):
    with app.app_context():
        migrations_path = os.path.join(os.path.dirname(__file__), "migrations")
        upgrade(migrations_path) # equivalent to running "flask db upgrade" every load

    if User.query.count() == 0:
        # Add temporary admin user
        password = gen_pass()
        add_user("root", "example@example.com", password, "admin")
        print("=================================================")
        print("ADMIN LOGIN CREDENTIALS (WILL ONLY BE SHOWN ONCE)")
        print("If you forget you'll need to manually add an")
        print("admin user to the database.")
        print("Username: root")
        print("Password:", password)
        print("=================================================")
    app.config['TEMPLATES_AUTO_RELOAD'] = True
    app.run(host='0.0.0.0', debug=debug)


from flask import Flask

from flask_sqlalchemy import SQLAlchemy
from flask_migrate import Migrate
from flask_login import LoginManager
from flask_admin import Admin
from flask_bootstrap import Bootstrap

from .config import Config
import pathlib

template_path = pathlib.Path(__file__).parent.absolute() / 'templates'
app = Flask(__name__, template_folder=str(template_path))
app.config.from_object(Config)

db = SQLAlchemy(app)
migrate = Migrate(app, db)
bootstrap = Bootstrap(app)

login = LoginManager(app)
login.login_view = 'login'

app.config['FLASK_ADMIN_SWATCH'] = 'cerulean'
admin = Admin(app, name='Admin Dashboard', template_mode='bootstrap3')

import pyfra.web.models




import pyfra.remote
import pyfra.shell
from functools import wraps, partial


def apt(r, packages):
    # install sudo if it's not installed; this is the case in some docker containers
    r.sh(f"sudo echo hi || {{ apt-get update; apt-get install sudo; }}; sudo apt-get update; sudo apt-get install -y {' '.join(packages)}", pyenv_version=None)


def ensure_supported(r):
    supported = [
        "Ubuntu 18", "Ubuntu 20", 
        "stretch"   # debian stretch
    ]

    @pyfra.remote.block
    def _f(r):
        print("Checking if", r, "is running a supported distro")

        assert any([
            ver in r.sh("lsb_release -d")
            for ver in supported
        ])

    _f(r)

## things to install


def install_pyenv(r, version="3.9.4"):
    if r.sh(f"pyenv shell {version} 2> /dev/null; python --version", no_venv=True, ignore_errors=True, pyenv_version=None, quiet=True).strip().split(" ")[-1] == version:
        return

    apt(r, [
        'build-essential',
        'curl',
        'git',
        'libbz2-dev',
        'libffi-dev',
        'liblzma-dev',
        'libncurses5-dev',
        'libncursesw5-dev',
        'libreadline-dev',
        'libsqlite3-dev',
        'libssl-dev',
        'make',
        'python3-openssl',
        'rsync',
        'tk-dev',
        'wget',
        'xz-utils',
        'zlib1g-dev',
    ])
    r.sh("curl https://pyenv.run | bash", ignore_errors=True, pyenv_version=None)

    payload = """
# pyfra-managed: pyenv stuff
export PYENV_ROOT="$HOME/.pyenv"
export PATH="$PYENV_ROOT/bin:$PATH"
eval "$(pyenv init --path)"
eval "$(pyenv init -)"
eval "$(pyenv virtualenv-init -)"
"""
    bashrc = r.sh("cat ~/.bashrc", pyenv_version=None)

    if "# pyfra-managed: pyenv stuff" not in bashrc:
        r.sh(f"echo {payload | pyfra.shell.quote} >> ~/.bashrc", pyenv_version=None)

    # install updater
    r.sh("git clone https://github.com/pyenv/pyenv-update.git $(pyenv root)/plugins/pyenv-update", ignore_errors=True, pyenv_version=None)
    r.sh("pyenv update", ignore_errors=True, pyenv_version=None)

    r.sh(f"pyenv install --verbose -s {version}", pyenv_version=None)

    # make sure the versions all check out
    assert r.sh(f"python --version", no_venv=True).strip().split(" ")[-1] == version
    assert r.sh(f"python3 --version", no_venv=True).strip().split(" ")[-1] == version
    assert version.rsplit('.', 1)[0] in r.sh("pip --version", no_venv=True)
    assert version.rsplit('.', 1)[0] in r.sh("pip3 --version", no_venv=True)
    
    r.sh("pip install virtualenv")
    r.sh("virtualenv --version")



"""Add user attributes

Revision ID: 1b1532ba9506
Revises: 1582bc5883ac
Create Date: 2021-04-22 18:42:10.038438

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1b1532ba9506'
down_revision = '1582bc5883ac'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('user', sa.Column('attributes', sa.String(length=512), nullable=False, server_default="{}"))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('user', 'attributes')
    # ### end Alembic commands ###



from flask_wtf import FlaskForm
from wtforms import StringField, PasswordField, BooleanField, SubmitField
from wtforms.validators import DataRequired, Email, EqualTo

class LoginForm(FlaskForm):
    name = StringField(validators=[DataRequired()])
    password = PasswordField(validators=[DataRequired()])
    remember_me = BooleanField()
    submit = SubmitField()

class ResetPasswordRequestForm(FlaskForm):
    email = StringField('Email', validators=[DataRequired(), Email()])
    submit = SubmitField('Request Password Reset')

class ResetPasswordForm(FlaskForm):
    password = PasswordField('New Password', validators=[DataRequired()])
    password2 = PasswordField(
        'Repeat New Password', validators=[DataRequired(), EqualTo('password')])
    submit = SubmitField('Change Password')



from colorama import init
init()

from .remote import *
from .web import *
from .shell import *
from .experiment import *


# fallback snippet from https://github.com/gruns/icecream
try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa


import errno
from hashlib import new
import os
import random
import re
import shlex
import shutil
import subprocess
import sys
import time
from typing import Union
import urllib
import json

from best_download import download_file
from colorama import Fore, Style
from natsort import natsorted
import pyfra.remote
import re


class ShellException(Exception): pass


__all__ = ['sh', 'copy', 'ls', 'rm', 'curl', 'quote']


def _wrap_command(x, no_venv=False, pyenv_version=None):
    bashrc_payload = r"""import sys,re; print(re.sub("If not running interactively.{,128}?esac", "", sys.stdin.read(), flags=re.DOTALL).replace('[ -z "$PS1" ] && return', ''))"""
    hdr = f"shopt -s expand_aliases; ctrlc() {{ echo Shell wrapper interrupted with C-c, raising error; exit 174; }}; trap ctrlc SIGINT; "
    hdr += f"eval \"$(cat ~/.bashrc | python3 -c {bashrc_payload | quote})\"  > /dev/null 2>&1; "
    hdr += "python() { python3 \"$@\"; };" # use python3 by default
    if pyenv_version is not None: hdr += f"pyenv shell {pyenv_version} || exit 1 > /dev/null 2>&1; "
    if not no_venv: hdr += "[ -f env/bin/activate ] && . env/bin/activate; "
    return hdr + x


def _process_remotepaths(host, cmd):
    candidates = re.findall(r"RemotePath\((.+?)\)", cmd)

    rempaths = []

    for c in candidates:
        ob = json.loads(c)
        rem = ob["remote"] if ob["remote"] is not None else "127.0.0.1"
        fname = ob["fname"]

        if rem != host:
            loc_fname = rem.replace(".", "_").replace("@", "_")+"_"+fname.split("/")[-1]
            if loc_fname.startswith("~/"): loc_fname = loc_fname[2:]
            if loc_fname.startswith("/"): loc_fname = loc_fname[1:]
            loc_fname = "~/.pyfra_remote_files/" + loc_fname

            copyerr = False
            try:
                frm = pyfra.remote.Remote(rem).path(fname)

                # we want to copy dirs into, but into doesnt work with files
                copy(frm, pyfra.remote.Remote(host).path(loc_fname), into=not frm.is_dir())
            except ShellException:
                # if this file doesn't exist, it's probably an implicit return
                copyerr = True

            rempaths.append((pyfra.remote.Remote(rem).path(fname), pyfra.remote.Remote(host).path(loc_fname), copyerr))

            cmd = cmd.replace(f"RemotePath({c})", loc_fname)
        else:
            cmd = cmd.replace(f"RemotePath({c})", fname)
    
    return cmd, rempaths


def _sh(cmd, quiet=False, wd=None, wrap=True, maxbuflen=1000000000, ignore_errors=False, no_venv=False, pyenv_version=None):
    if wrap: cmd = _wrap_command(cmd, no_venv=no_venv, pyenv_version=pyenv_version)

    if wd is None: wd = "~"

    cmd = f"cd {wd} > /dev/null 2>&1; {cmd}"

    p = subprocess.Popen(cmd, shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        executable="/bin/bash")
    
    ret = bytearray()
    while True:
        byte = p.stdout.read(1)

        if byte == b'':
            break
        if not quiet:
            sys.stdout.buffer.write(byte)
            sys.stdout.flush()

        if maxbuflen is None or len(ret) < maxbuflen:
            ret += bytearray(byte)
    
    p.communicate()
    if p.returncode == 174:
        raise KeyboardInterrupt()
    elif p.returncode != 0 and not ignore_errors:
        raise ShellException(p.returncode)

    return ret.decode("utf-8").replace("\r\n", "\n").strip()


def sh(cmd, quiet=False, wd=None, wrap=True, maxbuflen=1000000000, ignore_errors=False, no_venv=False, pyenv_version=None):
    """
    Runs commands as if it were in a local bash terminal.

    This function patches out the non-interactive detection in bashrc and sources it, activates virtualenvs and sets pyenv shell, handles interrupts correctly, and returns the text printed to stdout.

    Args:
        quiet (bool): If turned on, nothing is printed to stdout.
        wd (str): Working directory to run in. Defaults to ~
        wrap (bool): Magic for the bashrc, virtualenv, interrupt-handing, and pyenv stuff. Turn off to make this essentially os.system
        maxbuflen (int): Max number of bytes to save and return. Useful to prevent memory errors.
        ignore_errors (bool): If set, errors will be swallowed.
        no_venv (bool): If set, virtualenv will not be activated
        pyenv_version (str): Pyenv version to use. Will be silently ignored if not found.
    Returns:
        The standard output of the command, limited to maxbuflen bytes.
    """
    if wd is None: wd = os.getcwd()

    return _rsh("127.0.0.1", cmd, quiet, wd, wrap, maxbuflen, -1, ignore_errors, no_venv, pyenv_version)


def _rsh(host, cmd, quiet=False, wd=None, wrap=True, maxbuflen=1000000000, connection_timeout=10, ignore_errors=False, no_venv=False, pyenv_version=None, forward_keys=False):
    if host is None or host == "localhost": host = "127.0.0.1"

    # implicit-copy files from remote to local
    cmd, rempaths = _process_remotepaths(host, cmd)

    if not quiet:
        # display colored message
        host_style = Fore.GREEN+Style.BRIGHT
        sep_style = Style.NORMAL
        cmd_style = Fore.WHITE+Style.BRIGHT
        dir_style = Fore.BLUE+Style.BRIGHT
        hoststr = str(host)
        if wd is not None: 
            wd_display = wd
            if not wd.startswith("~/") and wd != '~':
                wd_display = os.path.join("~", wd)
        else:
            wd_display = "~"
        hoststr += f"{Style.RESET_ALL}:{dir_style}{wd_display}{Style.RESET_ALL}"
        cmd_fmt = cmd.strip().replace('\n', f'\n{ " " * (len(str(host)) + 3 + len(wd_display))}{sep_style}>{Style.RESET_ALL}{cmd_style} ')
        print(f"{Style.BRIGHT}{Fore.RED}*{Style.RESET_ALL} {host_style}{hoststr}{Style.RESET_ALL}{sep_style}$ {Style.RESET_ALL}{cmd_style}{cmd_fmt}{Style.RESET_ALL}")
    
    if host == "127.0.0.1":
        return _sh(cmd, quiet, wd, wrap, maxbuflen, ignore_errors, no_venv, pyenv_version)

    if wrap: cmd = _wrap_command(cmd, no_venv=no_venv, pyenv_version=pyenv_version)
    if wd: cmd = f"cd {wd}  > /dev/null 2>&1; {cmd}"
 
    ssh_cmd = "eval \"$(ssh-agent -s)\"; ssh-add ~/.ssh/id_rsa; ssh -A" if forward_keys else "ssh"
    ret = _sh(f"{ssh_cmd} -q -oConnectTimeout={connection_timeout} -oBatchMode=yes -oStrictHostKeyChecking=no -oUserKnownHostsFile=/dev/null -t {host} {shlex.quote(cmd)}", quiet=quiet, wrap=False, maxbuflen=maxbuflen, ignore_errors=ignore_errors, no_venv=no_venv)

    # implicit-copy files from local to remote
    for remf, locf, copyerr in rempaths:
        try:
            copy(locf, remf)
        except ShellException:
            # if it errored both before and after, something is wrong
            if copyerr:
                raise ShellException(f"implicit-copy file {remf}/{locf} (remote/local) was neither written to nor read from!")

    return ret

def copy(frm, to, quiet=False, connection_timeout=10, symlink_ok=True, into=True, exclude=[]) -> None:
    """
    Copies things from one place to another.

    Args:
        frm (str or RemotePath): Can be a string indicating a local path, a :class:`pyfra.remote.RemotePath`, or a URL.
        to (str or RemotePath): Can be a string indicating a local path or a :class:`pyfra.remote.RemotePath`.
        quiet (bool): Disables logging.
        connection_timeout (int): How long in seconds to give up after
        symlink_ok (bool): If frm and to are on the same machine, symlinks will be created instead of actually copying. Set to false to force copying.
        into (bool): If frm is a file, this has no effect. If frm is a directory, then into=True for frm="src" and to="dst" means "src/a" will get copied to "dst/src/a", whereas into=False means "src/a" will get copied to "dst/a".
    """

    # copy from url
    if isinstance(frm, str) and (frm.startswith("http://") or frm.startswith("https://")):
        if ":" in to:
            to_host, to_path = to.split(":")
            _rsh(to_host, f"curl {frm} --create-dirs -o {to_path}")
        else:
            wget(frm, to)
        return

    # get rsync strs and make sure frm and to are RemotePaths
    if isinstance(frm, pyfra.remote.RemotePath): 
        frm_str = frm.rsyncstr()
    else:
        frm_str = frm
        assert ":" not in frm_str
        frm = pyfra.remote.local.path(frm)

    if isinstance(to, pyfra.remote.RemotePath):
        to_str = to.rsyncstr()
    else:
        to_str = to
        assert ":" not in to_str
        to = pyfra.remote.local.path(to)
    
    # state tracking
    needs_set_kv = False
    if not to.remote._no_hash:
        with to.remote.no_hash():
            checksum = frm.quick_hash()
            new_hash = to.remote.update_hash("copy", to.fname, checksum)
            try:
                to.remote.get_kv(new_hash)
                # if already copied, then return
                to._set_cache("quick_hash", checksum) # set the checksum of the target file to avoid needing to calculate it again
                pyfra.remote._print_skip_msg(to.remote.envname, "copy", new_hash)

                return
            except KeyError:
                needs_set_kv = True

    # print info
    if not quiet: print(f"{Style.BRIGHT}{Fore.RED}*{Style.RESET_ALL} Copying {Style.BRIGHT}{frm_str} {Style.RESET_ALL}to {Style.BRIGHT}{to_str}{Style.RESET_ALL}")

    if frm_str[-1] == '/' and len(frm_str) > 1: frm_str = frm_str[:-1]
    if not into: frm_str += '/'

    if quiet:
        opts = "-e \"ssh -o StrictHostKeyChecking=no\" -arqL"
    else:
        opts = "-e \"ssh -o StrictHostKeyChecking=no\" -arL --info=progress2"
    
    for ex in exclude:
        opts += f" --exclude {ex | pyfra.shell.quote}"
    
    def symlink_frm(frm_str):
        # rsync behavior is to_str copy the contents of frm_str into to_str if frm_str ends with a /
        if frm_str[-1] == '/': frm_str += '*'
        # ln -s can't handle relative paths well! make absolute if not already
        if frm_str[0] != '/' and frm_str[0] != '~': frm_str = "$PWD/" + frm_str

        return frm_str

    if ":" in frm_str and ":" in to_str:
        frm_host, frm_path = frm_str.split(":")
        to_host, to_path = to_str.split(":")

        par_target = to_path.rsplit('/', 1)[0] if "/" in to_path else ""

        if to_host == frm_host:
            if symlink_ok:
                assert not exclude, "Cannot use exclude symlink"
                _rsh(frm_host, f"[ -d {frm_path} ] && mkdir -p {to_path}; ln -sf {symlink_frm(frm_path)} {to_path}", quiet=True)
            else:

                if par_target: _rsh(to_host, f"mkdir -p {par_target}", quiet=True)
                _rsh(frm_host, f"rsync {opts} {frm_path} {to_path}", quiet=True)
        else:
            rsync_cmd = f"rsync {opts} {frm_path} {to_str}"
                
            # make parent dir in terget if not exists
            if par_target: _rsh(to_host, f"mkdir -p {par_target}", quiet=True)

            sh(f"eval \"$(ssh-agent -s)\"; ssh-add ~/.ssh/id_rsa; ssh -q -oConnectTimeout={connection_timeout} -oBatchMode=yes -oStrictHostKeyChecking=no -oUserKnownHostsFile=/dev/null -A {frm_host} {rsync_cmd | quote}", wrap=False, quiet=True)
    else:
        # if to_str is host:path, then this gives us path; otherwise, it leaves it unchanged
        par_target = to_str.split(":")[-1]
        par_target = par_target.rsplit('/', 1)[0] if "/" in par_target else ""

        if symlink_ok and ":" not in frm_str and ":" not in to_str:
            assert not exclude, "Cannot use exclude symlink"
            sh(f"[ -d {frm_str} ] && mkdir -p {par_target}; ln -sf {symlink_frm(frm_str)} {to_str}", quiet=True)
        else:
            if ":" in to_str: _rsh(to_str.split(":")[0], f"mkdir -p {par_target}", quiet=True)
            sh((f"mkdir -p {par_target}; " if par_target and ":" in frm_str else "") + f"rsync {opts} {frm_str} {to_str}", wrap=False, quiet=True)
    
    # set value in key value store to flag as done
    if needs_set_kv:
        to.remote.set_kv(new_hash, None)
        to._set_cache("quick_hash", checksum) # set the checksum of the target file to avoid needing to calculate it again


def ls(x='.'):
    return list(natsorted([x + '/' + fn for fn in os.listdir(x)]))

def rm(x, no_exists_ok=True):
    # from https://stackoverflow.com/a/41789397
    if not os.path.exists(x) and no_exists_ok: return

    if os.path.isfile(x) or os.path.islink(x):
        os.remove(x)  # remove the file
    elif os.path.isdir(x):
        shutil.rmtree(x)  # remove dir and all contains
    else:
        raise ValueError("file {} is not a file or dir.".format(x))

def curl(url, max_tries=10, timeout=30): # TODO: add checksum option
    cooldown = 1
    for i in range(max_tries):
        try:
            response = urllib.request.urlopen(url, timeout=timeout)
            data = response.read()
        except urllib.error.URLError as e:
            if e.code == 429:
                time.sleep(cooldown)
                cooldown *= 1.5
                continue
            else:
                return None
        except:
            return None
        return data

# convenience function for shlex.quote
class _quote:
    def __ror__(self, other):
        return shlex.quote(other)
    
    def __call__(self, other):
        return shlex.quote(other)

quote = _quote()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyfra",
    version="0.3.0rc0",
    author="Leo Gao",
    author_email="lg@eleuther.ai",
    description="A framework for research code",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EleutherAI/pyfra",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'best_download',
        'flask',
        'flask-login',
        'flask-wtf',
        'flask-sqlalchemy',
        'flask-migrate',
        'flask-admin',
        'flask-bootstrap',
        'pyjwt',
        'sqlalchemy',
        'wtforms[email]',
        'ansi2html',
        'sqlitedict',
        'colorama',
        'parse',
        'natsort',
        'yaspin',
        'imohash',
    ]
)



from flask import Flask, request, render_template, render_template_string, json, jsonify, Response, flash, redirect, url_for, send_file

from werkzeug.urls import url_parse

from flask_login import current_user, login_user, login_required, logout_user

from .forms import LoginForm
from .forms import ResetPasswordRequestForm
from .forms import ResetPasswordForm

from .emailer import send_password_reset_email

from .app import app
from .app import db
from .app import admin

from .models import User

from flask_admin.contrib.sqla import ModelView

import collections
import pathlib

class PageRegistry:
    def __init__(self):
        self.registry = collections.defaultdict(list)
        self.i = 0

    def add_page(self, name, pretty_name, allowed_roles):
        for role in allowed_roles:
            self.registry[role].append((self.i, name, pretty_name))
        self.i += 1
    
    def get_pages(self, roles):
        ret = []
        for role in roles:
            for page in self.registry[role]:
                if page not in ret: ret.append(page)
        
        return [(n,p) for _,n,p in sorted(ret)]

registry = PageRegistry()


def add_user(name, email, password, roles=[]):
    newUser = User(email=email, name=name, roles=roles)

    newUser.set_password(password)

    db.session.add(newUser)        
    db.session.commit()


# Index and 404
# ===================================================================
@app.route('/')
@login_required
def index():
    roles = current_user.get_roles()
    pages = registry.get_pages(roles)
    return render_template('index.html', pages=pages)

@app.errorhandler(404)
def page_not_found(e):
    return render_template('404.html'), 404

# WTForms stuff
# ===================================================================
template_path = pathlib.Path(__file__).parent.absolute() / 'templates'

registry.add_page("admin/user", "Admin Dashboard", ["admin"])
registry.add_page("change_password", "Change Password", ["everyone"])

def register_page(name, pretty_name, form_class, callback, allowed_roles, redirect_index, has_form):
    if pretty_name is None: pretty_name = name

    registry.add_page(name, pretty_name, allowed_roles)

    @login_required
    def _fn():

        is_authorized = any([
            role in current_user.get_roles()
            for role in allowed_roles
        ])

        if not is_authorized:
            flash("You are not authorized to view this page.")
            return redirect(url_for('index'))

        if has_form:
            form = form_class()()

            html = ""
            if form.validate_on_submit():
                data = {
                    field.name: field.data
                    for field in form
                    if field.name not in ["submit", "csrf_token"]
                }

                ret = callback(data)
                html = ret if ret is not None else ""
                
                if redirect_index:
                    flash("Form successfully submitted")
                    return redirect(url_for('index'))
        else:
            html = callback({})
            form = None

        return render_template_string(open(template_path / 'form_template.html').read(), body=html, form=form, title=pretty_name, has_form=has_form)
    
    app.add_url_rule(f"/{name}", name, _fn, methods=['GET', 'POST'])


# Authentication Stuff
# ===================================================================
@app.route('/change_password/')
@login_required
def change_password():
    token = current_user.get_reset_password_token().decode()
    return redirect(f'/reset_password/{token}')

@app.route('/reset_password/<token>', methods=['GET', 'POST'])
def reset_password(token):
    user = User.verify_reset_password_token(token)
    if not user:
        return redirect(url_for('index'))
    form = ResetPasswordForm()
    if form.validate_on_submit():
        user.set_password(form.password.data)
        db.session.commit()
        flash('Your password has been reset.')
        return redirect(url_for('login'))
    return render_template('reset_password.html', form=form)

@app.route('/forgot-password', methods=['GET', 'POST'])
def resetPassword():
    if current_user.is_authenticated:
        return redirect(url_for('index'))

    form = ResetPasswordRequestForm()
    if form.validate_on_submit():
        user = User.query.filter_by(email=form.email.data).first()
        if user:
            token = user.get_reset_password_token()
            send_password_reset_email(app.config, user.email, user.name, token)
        flash('Check your email for the instructions to reset your password')
        return redirect(url_for('login'))
    return render_template('forgot_password.html',
                           title='Reset Password', form=form)

@app.route('/login', methods=['GET', 'POST'])
def login():
    if current_user.is_authenticated:
        return redirect(url_for('index'))

    form = LoginForm()
    if form.validate_on_submit():
        user = User.query.filter_by(name=form.name.data).first()
        if user is None or not user.check_password(form.password.data):
            flash('Invalid username or password')
            return redirect(url_for('login'))
        login_user(user, remember=form.remember_me.data)
        next_page = request.args.get('next')
        if not next_page or url_parse(next_page).netloc != '':
            next_page = url_for('index')
        return redirect(next_page)

    return render_template('login.html', title='Sign In', form=form)

@app.route('/logout')
def logout():
    logout_user()
    return redirect(url_for('login')) 


# Admin Section
# ===================================================================
class UserAdminView(ModelView):
    create_modal = True
    edit_modal = True
    can_export = True

    form_columns = ('name', 'email', "roles")
    column_exclude_list = ["password_hash"]
    column_searchable_list = ['name', 'email', "roles"]
    column_filters = ['name', 'email', "roles"]

    def is_accessible(self):
        return "admin" in current_user.get_roles()

    def inaccessible_callback(self, name, **kwargs):
        flash("You need to be an admin to view this page.")
        return redirect(url_for('index'))

admin.add_view(UserAdminView(User, db.session))
